#include "jserver.h"

/* plans:
 * turn the thread into a generic worker thread
 * have the to/from functions start these for incoming packets
 * maintain a cache of free ones, when one frees up it goes to the cache
 * each worker thread receives a spacket via the message port
 * 
 */

/* worker thread waiting pool */
#define SESSION_WAITERS 10
pth_msgport_t swaiters[SESSION_WAITERS];
int swaiters_init = 1;

/* spacket definition */
#define SPACKET_TO 1
#define SPACKET_FROM 2
#define SPACKET_START 3
#define SPACKET_END 4

typedef struct {
    pth_message_t head; /* the standard pth message header */
    jpacket p;
    session s;
    int type;
} _spacket, *spacket;

void *js_worker_main(void *arg);

/* return an open worker */
pth_msgport_t js_session_worker(void)
{
    pth_attr_t attr;
    thread t;

    pth_msgport_t mp;
    int i;

    /* scan for a waiting worker */
    for(i=0;i<SESSION_WAITERS; i++)
        if(swaiters[i] != NULL)
        {
            log_debug(ZONE,"worker fetch returning swaiters[%d] %X",i,swaiters[i]);
            return swaiters[i];
        }

    /* create new thread and msgport */
    mp = pth_msgport_create("js_worker");
    t = thread_new();
    t->data = (void *)mp;

    /* start new worker thread */
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, 2*etherx_stack_default);
    t->id = pth_spawn(attr, js_worker_main, (void *)t);
    pth_attr_destroy(attr);

    /* put it in the waiting pool */
    for(i=0;i<SESSION_WAITERS; i++)
        if(swaiters[i] == NULL)
        {
            swaiters[i] = mp;
            break;
        }
 
    log_debug(ZONE,"worker fetch returning new swaiters[%d] %X",i,mp);
    return mp;
}

/* send the spacket */
void js_spacket(int type, session s, jpacket p)
{
    static pth_msgport_t unknown_mp = NULL;
    spacket q;
 
    if(s == NULL)
        return;
 
    log_debug(ZONE,"spacket %d to session %X packet %X",type,s,p);
 
    if(unknown_mp == NULL)
        unknown_mp = pth_msgport_find("js_unknown");

    /* get the mem from wherever we can */ 
    if(p != NULL)
        q = pmalloc(p->p, sizeof(_spacket));
    else
        q = pmalloc(s->p, sizeof(_spacket));
    q->type = type;
    q->s = s;
    q->p = p;
 
    q->head.m_replyport = unknown_mp;
    pth_msgport_put(js_session_worker(), (pth_message_t *)q);
}

session js_session_new(jid owner, session_onSend send, void *arg)
{
    /* creats a new session, registers the resource for it */
    pool p;
    session s;
    jid uid;
    int i;

    if(owner == NULL || owner->resource == NULL) /* illegal */
        return NULL;

    /* hack to init the array the first time */
    if(swaiters_init)
    {
        for(i=0;i<SESSION_WAITERS;i++)
            swaiters[i] = NULL;
        swaiters_init = 0;
    }
 
    log_debug(ZONE,"session_create");
 
    /* create session thread */
    p = pool_new();
    s = pmalloc(p, sizeof(struct session_struct));
    s->p = p;

    /* save the send callback for the service */
    s->send = send;
    s->arg = arg;

    /* session identity */
    s->id = jid_new(p, jid_full(owner));
    uid = jid_new(p, jid_full(owner));
    jid_set(uid, NULL, JID_RESOURCE);
    s->uid = uid;
    s->res = pstrdup(p, owner->resource);
    s->u = js_user(owner->user);

    /* default settings */
    s->exit_flag = 0;
    s->roster = 0;
    s->priority = -1;
    s->presence = NULL;
    s->c_in = s->c_out = 0;
    s->m_out = s->m_in = s->m_end = NULL;

    /* remove any other session w/ this resource */
    js_session_end(js_session_get(s->u, owner->resource), "Replaced by new connection");

    /* make sure we're linked with the user */
    s->next = s->u->sessions;
    s->u->sessions = s;
    s->u->scount++;

    /* start it */
    js_spacket(SPACKET_START, s, NULL);

    return s;
}

void js_session_end(session s, char *reason)
{
    xmlnode x;
    jid to_id;
    session cur;

    if(s == NULL || s->exit_flag == 1 || reason == NULL)
        return;

    log_debug(ZONE,"end %d '%s'",s,reason);

    /* flag the session to exit ASAP */
    s->exit_flag = 1;

    /* make sure we're not the primary session */
    s->priority = -1;

    /* if the last known presence was available, update it */
    if(s->presence != NULL && j_strcmp(xmlnode_get_attrib(s->presence, "type"), "unavailable") != 0)
    {
        x = xmlnode_new_tag("presence");
        xmlnode_put_attrib(x,"type","unavailable");
        xmlnode_put_attrib(x,"from",jid_full(s->id));
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),reason,strlen(reason));
        xmlnode_free(s->presence);
        s->presence = x;
    }

    /* unlist ourselves */
    if(s == s->u->sessions)
    {
        s->u->sessions = s->next;
    }else{
        for(cur = s->u->sessions; cur->next != s; cur = cur->next);
        cur->next = s->next;
    }
    s->u->ref++; /* so it doesn't get freed */

    /* tell it to exit */
    js_spacket(SPACKET_END, s, NULL);
}

/* loop processing all the packets for this worker thread */
void js_session_process(pth_msgport_t mp)
{
    spacket q;
    mmaster master;

    while(1)
    {
        q = (spacket)pth_msgport_get(mp);
        if(q == NULL)
            break;

        log_debug(ZONE,"PROCESS %X type %d for session %X",mp,q->type,q->s);

        switch(q->type)
        {

            case SPACKET_START: /* start the session */
                master = js_mapi_master(P_SESSION);
                js_mapi_call(P_SESSION, master->l, NULL, q->s->u, q->s, NULL);
                q->s->started = time(NULL);
                break;

            case SPACKET_FROM: /* outgoing packets from the session */

                /* if this session is dead */
                if(q->s->exit_flag)
                {
                    xmlnode_free(q->p->x);
                    break;
                }

                /* at least we must have a valid packet */
                if(q->p->type == JPACKET_UNKNOWN)
                {
                    jutil_error(q->p->x,TERROR_BAD);
                    jpacket_reset(q->p);
                    js_session_to(q->s,q->p);
                    break;
                }

                log_debug(ZONE,"THREAD:SESSION:FROM received a message!");
                q->s->c_out++;

                /* make sure we have our from set correctly for outgoing crap */
                if(jid_cmp(q->p->from,q->s->uid) != 0 && jid_cmp(q->p->from,q->s->id) != 0)
                {
                    xmlnode_put_attrib(q->p->x,"from",jid_full(q->s->id));
                    q->p->from = jid_new(q->p->p,jid_full(q->s->id));
                }

                /* let the modules have their heyday */
                if(js_mapi_call(PS_OUT, q->s->m_out, q->p, q->s->u, q->s, jpacket_subtype(q->p)))
                    break;

                if(q->p->to == NULL)
                {
                    xmlnode_put_attrib(q->p->x,"to",q->p->from->server);
                    q->p->to = jid_new(q->p->p,q->p->from->server);
                }

                /* pass these to the general delivery function */
                js_deliver(q->p);

                break;

            case SPACKET_TO: /* incoming packets for the session */

                /* if this session is dead */
                if(q->s->exit_flag)
                {
                    if(q->p->type == JPACKET_MESSAGE)
                        js_deliver(q->p); /* re-deliver the messages only */
                    else
                      xmlnode_free(q->p->x);
                    break;
                }

                log_debug(ZONE,"THREAD:SESSION:TO received data from %s!",jid_full(q->p->from));
                q->s->c_in++;

               /* let the modules have their heyday */
                if(js_mapi_call(PS_IN, q->s->m_in, q->p, q->s->u, q->s, jpacket_subtype(q->p)))
                    break;

                /* we need to check again, s->exit_flag *could* have changed within the modules at some point */
                if(q->s->exit_flag)
                { /* save it if it was a message, and sk'daddle */
                    if(q->p->type == JPACKET_MESSAGE)
                        js_deliver(q->p);
                    else
                        xmlnode_free(q->p->x);
                    break;
                }

                /* deliver outgoing for this session to the onSend event passed when this session was created */
                (q->s->send)(q->s, q->p, q->s->arg);

                break;

            case SPACKET_END: /* buh-bye */

                log_debug(ZONE,"THREAD:SESSION exiting");
                q->s->u->scount--;

                /* make sure the service knows the session is gone */
                (q->s->send)(q->s, NULL, q->s->arg);

                /* let the modules have their heyday */
                js_mapi_call(PS_END, q->s->m_end, NULL, q->s->u, q->s, 0);

                xmlnode_free(q->s->presence);

                q->s->u->ref--; /* let the user struct go  */
                break;

            default:
        }
    }
}

void *js_worker_main(void *arg)
{
    thread t = (thread)arg;
    pth_msgport_t mp = (pth_msgport_t)(t->data);
    pth_event_t mpevt;
    int i;

    log_debug(ZONE,"THREAD:WORKER %X starting",mp);

    mpevt = pth_event(PTH_EVENT_MSG,mp);

    while(1)
    {
        log_debug(ZONE,"WORKER(%X)->pth",mp);
        pth_wait(mpevt);
        log_debug(ZONE,"pth->WORKER(%X)",mp);

        /* take us out of the waiting pool */
        for(i=0;i<SESSION_WAITERS; i++)
            if(swaiters[i] == mp)
                swaiters[i] = NULL;

        /* process the waiting packets */
        js_session_process(mp);

        /* scan the waiting pool */
        for(i=0;i<SESSION_WAITERS && swaiters[i] != NULL; i++);

        /* no room for us, die */
        if(i+1 == SESSION_WAITERS)
            break;

        log_debug(ZONE,"swaiters[%d] is now %X",i,mp);
        swaiters[i] = mp;
    }

    log_debug(ZONE,"THREAD:WORKER %X exiting",mp);

    pool_free(t->p);
    pth_event_free(mpevt,PTH_FREE_ALL);
    pth_msgport_destroy(mp);
}

session js_session_get(udata user, char *res)
{
    session cur;
 
    if(user == NULL || res == NULL)
        return NULL;
 
    for(cur = user->sessions; cur != NULL; cur = cur->next)
        if(j_strcmp(res, cur->res) == 0)
            return cur;
 
    return NULL;
}

session js_session_primary(udata user)
{
    session cur, top;
 
    if(user == NULL || user->sessions == NULL)
        return NULL;
 
    /* find primary session */
    top = user->sessions;
    for(cur = top; cur != NULL; cur = cur->next)
        if(cur->priority > top->priority)
            top = cur;
 
    if(top->priority >= 0)
        return top;
    else
        return NULL;
}

void js_session_to(session s, jpacket p)
{
    js_spacket(SPACKET_TO,s,p);
}

void js_session_from(session s, jpacket p)
{
    js_spacket(SPACKET_FROM,s,p);
}

